unit Equations;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

const                               // EXAMPLES:
  eqstrFraction    = 'fraq';        //  fraq(a, b)
  eqstrRoot        = 'root';        //  root([a], x)
  eqstrUpperIndex  = 'up';          //  up(value, index)
  eqstrLowerIndex  = 'down';        //  down(value, index)
  eqstrBothIndex   = 'updown';      //  updown(value, up-index, dn-index)
  eqstrSymbol      = 'symb';        //  symb(str)     - 'Symbol' font
  eqstrText        = 'text';        //  text(str)     - standard font
  eqstrEquation    = 'eq';          //  eq(a,b,c,...)
  eqstrMatrix      = 'matrix';      //  matrix(rows,cols,a11,a12...,a21,a22...)
  eqstrScale       = 'scale';       //  scale(0.5, text(aaa))

type
  TDrawEquation = class (TGraphicControl)
  private
    fDescription: TStringList;
    fImg: TMetafile;
    procedure setDescription(value:TStringList);
    procedure UpdateImg;
    procedure UpdateEvent(Sender:TObject);
  protected
    procedure Resize; override;
    procedure Paint; override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
  published
    property Description:TStringList read fDescription write setDescription;
    property Font;
  end;


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('PJank', [TDrawEquation]);
end;

{ TDrawEquation }

constructor TDrawEquation.Create(AOwner: TComponent);
begin
  inherited;
  fDescription:= TStringList.Create;
  fDescription.OnChange:= UpdateEvent;
  fImg:= TMetafile.Create;
  Width:= 50;
  Height:= 20;
  ParentFont:= False;
  Font.OnChange:= UpdateEvent;
end;

destructor TDrawEquation.Destroy;
begin
  inherited;
  fImg.Free;
  fDescription.Free;
end;

procedure TDrawEquation.Paint;
begin
  Canvas.Draw(0,0,fImg);
end;

procedure TDrawEquation.Resize;
begin
  inherited;
  UpdateImg;
end;

procedure TDrawEquation.setDescription(value: TStringList);
begin
  fDescription.Assign(value);
  UpdateImg;
end;

procedure TDrawEquation.UpdateEvent(Sender: TObject);
begin
  UpdateImg;
end;

procedure TDrawEquation.UpdateImg;
var mc: TMetafileCanvas;

  function fix(const s:string):string;
  begin
    Result:= s;
    Result:= StringReplace(Result, #10, '', [rfReplaceAll]);
    Result:= StringReplace(Result, #13, '', [rfReplaceAll]);
  end;

  function DRAW(str:string; r:TRect; scale:Single; sizeonly:Boolean):Integer;
  var fun:string;
      idx: array of Integer;
      startpos,endpos: Integer;
      i,x,y,x1,x2,a,b,w,h,top,bot: Integer;
      xx: array of Integer;
      rr: TRect;
      s: string;

    procedure PARSE;
    var pos,level: Integer;
        special: Boolean;
      procedure newpos(pos:Integer);
      begin
        SetLength(idx, Length(idx)+1);
        idx[High(idx)]:= pos;
      end;
    begin
      startpos:= -1;
      SetLength(idx, 0);
      level:= 0;
      special:= False;
      for pos:= 1 to Length(str) do
        if special then special:= False else
        case str[pos] of
          '\': special:= True;
          '(': begin
                 if level=0 then begin
                   startpos:= pos;
                   newpos(startpos);
                 end;
                 Inc(level);
               end;
          ')': begin
                 Dec(level);
                 if level=0 then begin
                   endpos:= pos;
                   newpos(endpos);
                   Break;
                 end;
               end;
          ',': if level=1 then newpos(pos);
          else if level=0 then fun:= fun+str[pos];
        end;
      if (startpos<0) or (level>0) then Abort;
      fun:= TrimLeft(fun);
    end;  // ...PARSE

    function param(i:Integer):string;
    var a,b: Integer;
    begin
      if i+1>High(idx) then Abort;
      a:= idx[i];
      b:= idx[i+1];
      Result:= Copy(str, a+1, b-a-1);
    end;

    procedure drawText;
    begin
      s:= Copy(str, startpos+1, endpos-startpos-1);
      s:= StringReplace(s, '\', '', [rfReplaceAll]);
      mc.Font.Height:= trunc((r.Bottom-r.Top)*scale);
      Result:= mc.TextWidth(s);
      if not sizeonly then begin
        h:= mc.TextHeight(s);
        mc.TextRect(r, r.Left, r.Top+(r.Bottom-r.Top-h) div 2, s);
      end;
    end;

  begin
    if Trim(str)='' then begin
      Result:= 1;
      Exit;
    end;
    Result:= 0;
    PARSE;
    if fun=eqstrFraction then begin        // fraction
      y:= (r.Bottom-r.Top) div 2;
      x1:= draw(param(0), Rect(0, r.Top, 0, r.Top+y-1), scale, True);
      x2:= draw(param(1), Rect(0, r.Top+y+1,   0, r.Bottom), scale, True);
      if x2>x1 then Result:= x2 else Result:= x1;
      if not sizeonly then begin
        mc.Pen.Width:= 1 + y div 20;
        mc.MoveTo(r.Left, r.Top+y);
        mc.LineTo(r.Left+2+Result, r.Top+y);
        draw(param(0), Rect(r.Left+1+(Result-x1) div 2, r.Top, r.Right-1, r.Top+y-1), scale, False);
        draw(param(1), Rect(r.Left+1+(Result-x2) div 2, r.Top+y+1,   r.Right-1, r.Bottom), scale, False);
      end;
    end else
    if fun=eqstrRoot then begin            // root
      a:= (r.Bottom-r.Top) div 2;
      b:= 1 + a div 20;
      mc.Pen.Width:= b;
      Result:= 5*b;
      if High(idx)>1 then begin
        x1:= draw(param(0), r, scale, True);
        if x1>Result then Result:= x1;
      end;
      rr:= Rect(r.Left+Result+2*b, r.Top+2*b, r.Right, r.Bottom);
      x2:= draw(param(High(idx)-1), rr, scale, True);
      if not sizeonly then begin
        mc.MoveTo(r.Left+Result-5*b, a+b);
        mc.LineTo(r.Left+Result-2*b, r.Bottom);
        mc.LineTo(r.Left+Result, r.Top);
        mc.LineTo(r.Left+Result+x2+3*b, r.Top);
        mc.LineTo(r.Left+Result+x2+3*b, r.Top+2*b);
        if High(idx)>1 then begin
          r.Right:= r.Left+Result-b;
          r.Bottom:= r.Top+a-b;
          draw(param(0), r, scale, False);
        end;
        draw(param(High(idx)-1), rr, scale, False);
      end;
      Inc(Result, x2+3*b);
    end else
    if fun=eqstrUpperIndex then begin      // upper index
      Result:= draw(param(0), r, scale, True);
      y:= (r.Bottom-r.Top) div 2 -2;
      x1:= draw(param(1), Rect(0, r.Top, 0, r.Bottom-y), scale, True);
      if not sizeonly then begin
        draw(param(0), r, scale, False);
        draw(param(1), Rect(r.Left+Result, r.Top, r.Right, r.Bottom-y), scale, False);
      end;
      Inc(Result,x1);
    end else
    if fun=eqstrLowerIndex then begin      // lower index
      Result:= draw(param(0), r, scale, True);
      y:= (r.Bottom-r.Top) div 2 -2;
      x1:= draw(param(1), Rect(0, r.Top+y, 0, r.Bottom), scale, True);
      if not sizeonly then begin
        draw(param(0), r, scale, False);
        draw(param(1), Rect(r.Left+Result, r.Top+y, r.Right, r.Bottom), scale, False);
      end;
      Inc(Result,x1);
    end else
    if fun=eqstrBothIndex then begin       // upper/lower index
      Result:= draw(param(0), r, scale, True);
      y:= (r.Bottom-r.Top) div 2 -1;
      x1:= draw(param(1), Rect(0, r.Top, 0, r.Bottom-y), scale, True);
      x2:= draw(param(2), Rect(0, r.Top+y, 0, r.Bottom), scale, True);
      if not sizeonly then begin
        draw(param(1), Rect(r.Left+Result, r.Top, r.Right, r.Bottom-y), scale, False);
        draw(param(2), Rect(r.Left+Result, r.Top+y, r.Right, r.Bottom), scale, False);
      end;
      if x2>x1 then Inc(Result,x2) else Inc(Result,x1);
    end else
    if fun=eqstrSymbol then begin          // symbol
      mc.Font.Name:= 'Symbol';
      mc.Font.Charset:= SYMBOL_CHARSET;
      drawText;
    end else
    if fun=eqstrText then begin            // text
      mc.Font:= Self.Font;
      drawText;
    end else
    if fun=eqstrEquation then begin        // mix
      for i:= 0 to High(idx)-1 do begin
        w:= draw(param(i), r, scale, sizeonly);
        Inc(Result, w);
        Inc(r.Left, w);
      end;
    end else
    if fun=eqstrMatrix then begin          // matrix
      a:= Round((r.Bottom-r.Top)*scale/2);
      b:= (r.Bottom-r.Top) div 2;
      top:= b-a;
      bot:= b+a;
      mc.Pen.Width:= 1 + (bot-top) div 100;
      w:= 1+mc.Pen.Width div 2;
      if not sizeonly then begin
        mc.MoveTo(r.Left+4*w, top+w);
        mc.LineTo(r.Left+w, top+w);
        mc.LineTo(r.Left+w, bot-w);
        mc.LineTo(r.Left+4*w, bot-w);
      end;
      Inc(Result, 4*w);
      try
        a:= StrToInt(param(0)); //rows
        b:= StrToInt(param(1)); //cols
      except
        on EConvertError do begin Abort; Exit; end;
      end;
      h:= (bot-top) div a;
      SetLength(xx, a+1);
      for x:= 1 to b do begin
        xx[0]:= 0;
        rr:= Rect(0, 0, 0, h);
        for y:= 1 to a do begin
          s:= param(b*(y-1)+x+1);
          xx[y]:= draw(s, rr, 1, True);
          if xx[y]>xx[0] then xx[0]:= xx[y];
        end;
        if not sizeonly then
          for y:= 1 to a do begin
            s:= param(b*(y-1)+x+1);
            rr.Left:= r.Left+Result+2 + (xx[0]-xx[y]) div 2;
            rr.Right:= r.Right;
            rr.Top:= top+h*(y-1);
            rr.Bottom:= top+h*y;
            draw(s, rr, 1, False);
          end;
        Inc(Result, xx[0]+4);
      end;
      mc.Pen.Width:= 1 + (bot-top) div 100;
      Inc(Result, 4*w);
      if not sizeonly then begin
        r.Right:= r.Left+Result;
        mc.MoveTo(r.Right-4*w, top+w);
        mc.LineTo(r.Right-w, top+w);
        mc.LineTo(r.Right-w, bot-w);
        mc.LineTo(r.Right-4*w, bot-w);
      end;
    end else
    if fun=eqstrScale then begin               // scale
      s:= param(0);
      s:= StringReplace(s, '.', DecimalSeparator, []);
      try scale:= scale*StrToFloat(s);
      except on EConvertError do Abort;
      end;
      Result:= draw(param(1), r, scale, sizeonly);
    end
    else Abort;
  end;  // ...DRAW

begin
  fImg.Clear;
  fImg.Width:= Width;
  fImg.Height:= Height;
  mc:= TMetafileCanvas.Create(fImg, 0);
  try
    mc.Font:= Font;
    mc.Brush.Style:= bsClear;
    mc.Pen.Color:= Font.Color;
    try
      draw(fix(fDescription.Text), Rect(0,0,Width,Height), 1.0, False);
    except
      on EAbort do;
    end;
  finally
    mc.Free;
    Invalidate;
  end;
end;

end.
