unit PJActiveBtn;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls;

type
  TPJActiveBtn = class;
  TPJBtnPressEvent = procedure(Sender:TPJActiveBtn; Button:TMouseButton; counter:Integer) of object;
  TPJActiveBtn = class(TGraphicControl)
  private
    fTimer: TTimer;
    fButton: TMouseButton;
    fInterval: Cardinal;
    fImgNormal: TBitmap;
    fImgActive: TBitmap;
    fImgDown: TBitmap;
    fActive: Boolean;
    fDown: Boolean;
    fStretch: Boolean;
    fAutoSize: Boolean;
    fTransparent: Boolean;
    fOnPress: TPJBtnPressEvent;
    fOnRelease: TNotifyEvent;
    fOnActivate: TNotifyEvent;
    fOnDeactivate: TNotifyEvent;
    procedure SetImgNormal(img:TBitmap);
    procedure SetImgActive(img:TBitmap);
    procedure SetImgDown(img:TBitmap);
    procedure SetStretch(value:Boolean);
    procedure SetAutoSize(value:Boolean);
    procedure SetTransparent(value:Boolean);
    procedure SetInterval(value:Cardinal);
    procedure cmMouseEnter(var M:TMessage); message CM_MOUSEENTER;
    procedure cmMouseLeave(var M:TMessage); message CM_MOUSELEAVE;
    procedure DoPress(Sender:TObject);
    procedure SetActive(value:Boolean);
  protected
    procedure Paint; override;
    procedure MouseDown(Button:TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseUp(Button:TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseMove(Shift:TShiftState; X,Y:Integer); override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure Press;
  published
    property OnPress:TPJBtnPressEvent read fOnPress write fOnPress;
    property OnRelease:TNotifyEvent read fOnRelease write fOnRelease;
    property OnActivate:TNotifyEvent read fOnActivate write fOnActivate;
    property OnDeactivate:TNotifyEvent read fOnDeactivate write fOnDeactivate;
    property Interval:Cardinal read fInterval write SetInterval;
    property ImageNormal:TBitmap read fImgNormal write SetImgNormal;
    property ImageActive:TBitmap read fImgActive write SetImgActive;
    property ImageDown:TBitmap read fImgDown write SetImgDown;
    property Stretch:Boolean read fStretch write SetStretch;
    property AutoSize:Boolean read fAutoSize write SetAutoSize default True;
    property Transparent:Boolean read fTransparent write SetTransparent;
    property Enabled;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('PJank', [TPJActiveBtn]);
end;

{ TPJActiveBtn }

constructor TPJActiveBtn.Create(AOwner: TComponent);
begin
  inherited;
  ControlStyle:= ControlStyle + [csOpaque];
  Width:= 80;
  Height:= 26;
  Cursor:= crHandPoint;
  fOnPress:= nil;
  fOnRelease:= nil;
  fTimer:= TTimer.Create(Self);
  fTimer.Enabled:= False;
  fTimer.OnTimer:= DoPress;
  fImgNormal:= TBitmap.Create;
  fImgActive:= TBitmap.Create;
  fImgDown:= TBitmap.Create;
  fActive:= False;
  fDown:= False;
  fAutoSize:= True;
  fStretch:= False;
  fTransparent:= False;
  Interval:= 0;
end;

destructor TPJActiveBtn.Destroy;
begin
  fTimer.Free;
  fImgNormal.Free;
  fImgActive.Free;
  fImgDown.Free;
  inherited;
end;

procedure TPJActiveBtn.cmMouseEnter(var M: TMessage);
begin
  inherited;
  SetActive(True);
  Invalidate;
end;

procedure TPJActiveBtn.cmMouseLeave(var M: TMessage);
begin
  inherited;
  SetActive(False);
  Invalidate;
end;

procedure TPJActiveBtn.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  SetActive(True);
  fDown:= True;
  fButton:= Button;
  Repaint;
  if fInterval<>0 then begin
    DoPress(nil);
    fTimer.Enabled:= True;
  end;
end;

procedure TPJActiveBtn.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  fTimer.Enabled:= False;
  fDown:= False;
  Invalidate;
  if fInterval=0 then DoPress(nil);
  if Assigned(fOnRelease) then fOnRelease(Self);
end;

procedure TPJActiveBtn.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
end;

procedure TPJActiveBtn.Paint;
var img: TBitmap;
begin
  if Enabled and fActive and fDown and not fImgDown.Empty
   then img:= fImgDown else
  if Enabled and fActive and not fImgActive.Empty
   then img:= fImgActive
   else img:= fImgNormal;
  img.Transparent:= fTransparent;
  img.TransparentMode:= tmAuto;
  if img.Empty
   then Canvas.TextRect(ClientRect, 2, 2, 'No Image!')
   else if fStretch
         then Canvas.StretchDraw(ClientRect, img)
         else Canvas.Draw(0,0,img);
end;

procedure TPJActiveBtn.SetImgActive(img: TBitmap);
begin
  fImgActive.Assign(img);
  Invalidate;
  if fAutoSize then SetBounds(Left, Top, img.Width, img.Height);
end;

procedure TPJActiveBtn.SetImgDown(img: TBitmap);
begin
  fImgDown.Assign(img);
  Invalidate;
  if fAutoSize then SetBounds(Left, Top, img.Width, img.Height);
end;

procedure TPJActiveBtn.SetImgNormal(img: TBitmap);
begin
  fImgNormal.Assign(img);
  Invalidate;
  if fAutoSize then SetBounds(Left, Top, img.Width, img.Height);
end;

procedure TPJActiveBtn.SetAutoSize(value: Boolean);
begin
  fAutoSize:= value;
  Invalidate;
  if fAutoSize then begin
    fStretch:= False;
    SetBounds(Left, Top, fImgNormal.Width, fImgNormal.Height);
  end;
end;

procedure TPJActiveBtn.SetStretch(value: Boolean);
begin
  fStretch:= value;
  if fStretch then begin
    fAutoSize:= False;
  end;
  Invalidate;
end;


procedure TPJActiveBtn.DoPress(Sender: TObject);
begin
  if (Sender=nil) or (fActive and fDown) then begin
    if Sender=nil then fTimer.Tag:= 0;
    fTimer.Tag:= fTimer.Tag+1;
    if Assigned(fOnPress) then fOnPress(Self, fButton, fTimer.Tag);
  end;
end;

procedure TPJActiveBtn.SetTransparent(value: Boolean);
begin
  fTransparent:= value;
  Invalidate;
end;

procedure TPJActiveBtn.SetActive(value: Boolean);
begin
  if fActive<>value then begin
    fActive:= value;
    if fActive then begin
      if Assigned(fOnActivate) then fOnActivate(Self)
    end
    else begin
      if Assigned(fOnDeactivate) then fOnDeactivate(Self);
    end;
  end;
end;

procedure TPJActiveBtn.SetInterval(value: Cardinal);
begin
  fInterval:= value;
  fTimer.Interval:= fInterval;
end;

procedure TPJActiveBtn.Press;
begin
  if Assigned(fOnPress) then fOnPress(Self, mbLeft, 1);
end;

end.
