unit PJSpin;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls;

type
  TPJSpinColors = class(TPersistent)
  private
    fOwner: TControl;
    fBtn: TColor;
    fText: TColor;
    fBack: TColor;
    fActiveBtn: TColor;
    fActiveText: TColor;
    fActiveBack: TColor;
    function Check(const old,new:TColor):TColor;
    procedure SetBtn(c:TColor);
    procedure SetBack(c:TColor);
    procedure SetText(c:TColor);
    procedure SetActiveBtn(c:TColor);
    procedure SetActiveBack(c:TColor);
    procedure SetActiveText(c:TColor);
  published
    property Btn:TColor read fBtn write SetBtn                      default clSilver;
    property Text:TColor read fText write SetText                   default clSilver;
    property Back:TColor read fBack write SetBack                   default clBlack;
    property ActiveBtn:TColor read fActiveBtn write SetActiveBtn    default clYellow;
    property ActiveText:TColor read fActiveText write SetActiveText default clYellow;
    property ActiveBack:TColor read fActiveBack write SetActiveBack default clBlack;
  public
    constructor Create(owner:TControl);
  end;

  TPJSpin = class(TCustomControl)
  private
    fValue: Single;
    fValueMin: Single;
    fValueMax: Single;
    fWrap: Boolean;
    fDigits: Integer;
    fColors: TPJSpinColors;
    fBtnsWidth: Integer;
    fMargin: Integer;
    fOnChange: TNotifyEvent;
    fOnBtnUp: TNotifyEvent;
    fOnBtnDn: TNotifyEvent;
    procedure SetValue(newValue:Single);
    function GetIntValue:Integer;
    procedure SetIntValue(newValue:Integer);
    procedure SetValueMin(newValueMin:Single);
    procedure SetValueMax(newValueMax:Single);
    procedure SetDigits(newDigits:Integer);
    procedure SetBtnsWidth(newWidth:Integer);
    procedure SetMargin(newMargin:Integer);
  private
    fMouseIn: Boolean;
    fBtnActive: Integer;
    fBtnDown: Integer;
    fRects: record Text,Btns,BtnUp,BtnDn,BtnUD:TRect end;
    fTimer: TTimer;
    fStep: Single;
    fLastDragPos: Integer;
    procedure TimerProc(Sender:TObject);
    procedure UpdateSize;
    procedure cmMouseEnter(var M:TMessage); message CM_MOUSEENTER;
    procedure cmMouseLeave(var M:TMessage); message CM_MOUSELEAVE;
    procedure wmSize(var M:TWMSize); message WM_SIZE;
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure CheckMousePos(X,Y:Integer);
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
  published  // published properties
    property Value:Single read fValue write SetValue;
    property IntValue:Integer read GetIntValue write SetIntValue;
    property ValueMin:Single read fValueMin write SetValueMin;
    property ValueMax:Single read fValueMax write SetValueMax;
    property Wrap:Boolean read fWrap write fWrap;
    property Digits:Integer read fDigits write SetDigits;
    property Colors:TPJSpinColors read fColors write fColors;
    property BtnsWidth:Integer read fBtnsWidth write SetBtnsWidth default 8;
    property Margin:Integer read fMargin write SetMargin default 2;
    property Font;
    property Enabled;
  published  // events
    property OnChange:TNotifyEvent read fOnChange write fOnChange;
    property OnBtnUp:TNotifyEvent read fOnBtnUp write fOnBtnUp;
    property OnBtnDn:TNotifyEvent read fOnBtnDn write fOnBtnDn;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('PJank', [TPJSpin]);
end;

{=============================================================================}



{ TPJSpinColors }
{=============================================================================}

constructor TPJSpinColors.Create(owner: TControl);
begin
  fOwner:= owner;
  fBtn:= clSilver;
  fText:= clSilver;
  fBack:= clBlack;
  fActiveBtn:= clYellow;
  fActiveText:= clYellow;
  fActiveBack:= clBlack;
end;

function TPJSpinColors.Check(const old, new: TColor): TColor;
begin
  if old<>new then fOwner.Invalidate;
  Result:= new;
end;

procedure TPJSpinColors.SetBtn(c: TColor);
begin
  fBtn:= Check(fBtn, c);
end;

procedure TPJSpinColors.SetBack(c: TColor);
begin
  fBack:= Check(fBack, c);
end;

procedure TPJSpinColors.SetText(c: TColor);
begin
  fText:= Check(fText, c);
end;

procedure TPJSpinColors.SetActiveBtn(c: TColor);
begin
  fActiveBtn:= Check(fActiveBtn, c);
end;

procedure TPJSpinColors.SetActiveBack(c: TColor);
begin
  fActiveBack:= Check(fActiveBack, c);
end;

procedure TPJSpinColors.SetActiveText(c: TColor);
begin
  fActiveText:= Check(fActiveText, c);
end;


{ TPJSpin }
{=============================================================================}

constructor TPJSpin.Create(AOwner: TComponent);
begin
  fColors:= TPJSpinColors.Create(Self);
  inherited Create(AOwner);
  ControlStyle:= ControlStyle + [csOpaque];
  with Font do begin
    Size:= 12;
    Name:= 'Arial Black';
    Style:= [];
  end;
  Width:= 60;
  Height:= 16;
  fBtnsWidth:= 8;
  fMargin:= 2;

  fValueMin:= 0;
  fValueMax:= 100;
  fValue:= 0;
  fWrap:= False;
  fDigits:= 0;
  fStep:= 1;

  fOnChange:= nil;
  fOnBtnUp:= nil;
  fOnBtnDn:= nil;

  fMouseIn:= False;
  fBtnActive:= 0;
  fBtnDown:= 0;
  fTimer:= nil;
end;

destructor TPJSpin.Destroy;
begin
  if (fBtnDown=2) then ReleaseCapture;
  fTimer.Free;
  inherited Destroy;
  fColors.Free;
end;

//------------------------------------- properties --------

procedure TPJSpin.SetValue(newValue: Single);
begin
  if newValue>fValueMax then
    if fWrap
     then newValue:= fValueMin + (newValue-fValueMax)
     else newValue:= fValueMax;
  if newValue<fValueMin then
    if fWrap
     then newValue:= fValueMax - (fValueMin-newValue)
     else newValue:= fValueMin;
  if newValue<>fValue then begin
    fValue:= newValue;
    Invalidate;
    if Assigned(fOnChange) then fOnChange(Self);
  end;
end;

function TPJSpin.GetIntValue: Integer;
begin
  Result:= Round(Value);
end;

procedure TPJSpin.SetIntValue(newValue: Integer);
begin
  Value:= newValue;
end;

procedure TPJSpin.SetValueMax(newValueMax: Single);
begin
  if newValueMax<>fValueMax then begin
    fValueMax:= newValueMax;
    if fValueMax<fValueMin then fValueMin:= fValueMax;
    SetValue(fValue);
  end;
end;

procedure TPJSpin.SetValueMin(newValueMin: Single);
begin
  if newValueMin<>fValueMin then begin
    fValueMin:= newValueMin;
    if fValueMin>fValueMax then fValueMax:= fValueMin;
    SetValue(fValue);
  end;
end;

procedure TPJSpin.SetDigits(newDigits: Integer);
begin
  if newDigits<0 then newDigits:= 0;
  if newDigits>5 then newDigits:= 5;
  if newDigits<>fDigits then begin
    fDigits:= newDigits;
    fStep:= 1;
    while newDigits>0 do begin
      fStep:= fStep/10;
      Dec(newDigits);
    end;
    Invalidate;
  end;
end;

procedure TPJSpin.SetBtnsWidth(newWidth: Integer);
begin
  if newWidth<>fBtnsWidth then begin
    fBtnsWidth:= newWidth;
    UpdateSize;
    Invalidate;
  end;
end;

procedure TPJSpin.SetMargin(newMargin: Integer);
begin
  if newMargin<>fMargin then begin
    fMargin:= newMargin;
    UpdateSize;
    Invalidate;
  end;
end;

//------------------------------------- resizing ----------

procedure TPJSpin.UpdateSize;
var a: Integer;
begin
  with fRects do begin
    Text:= Rect(1, 1, Width-2-fBtnsWidth-fMargin, Height-2);
    Btns:= Rect(Width-2-fBtnsWidth, 1, Width-2, Height-2);
    with Btns do begin
      a:= (Bottom-Top+1) div 2;
      BtnUp:= Rect(Left, Top, Right, a-1);
      BtnDn:= Rect(Left, Height-a, Right, Bottom);
      BtnUD:= Rect(Left, BtnUp.Bottom+1, Right, BtnDn.Top-1);
    end;
  end;
end;

procedure TPJSpin.wmSize(var M: TWMSize);
begin
  UpdateSize;
end;

//------------------------------------- mouse handling ----
procedure TPJSpin.cmMouseEnter(var M: TMessage);
begin
  inherited;
  fMouseIn:= True;
  Repaint;
end;

procedure TPJSpin.cmMouseLeave(var M: TMessage);
begin
  inherited;
  fMouseIn:= False;
  Invalidate;
end;

procedure TPJSpin.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  if (fBtnDown=2) then begin
    Value:= Value + (fLastDragPos-Y)*fStep;
    fLastDragPos:= Y;
  end
  else begin
    CheckMousePos(X,Y);
    if (fBtnDown<>0) then
      fTimer.Enabled:= (fBtnActive=fBtnDown);
  end;
end;

procedure TPJSpin.CheckMousePos(X, Y: Integer);
var p: TPoint;
    b: Integer;
begin
  p:= Point(X, Y);
  if not fMouseIn then begin
    fMouseIn:= True;
    Invalidate;
  end;
  if PtInRect(fRects.Btns, p) then begin
    if PtInRect(fRects.BtnUp, p) then b:= 1 else
    if PtInRect(fRects.BtnDn, p) then b:= 3
    else b:= 2;
  end
  else b:= 0;
  if (b<>fBtnActive) then begin
    fBtnActive:= b;
    if fBtnActive=0
     then Cursor:= crDefault
     else Cursor:= crHandPoint;
    Repaint;
  end;
end;

procedure TPJSpin.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  CheckMousePos(X,Y);
  fBtnDown:= fBtnActive;
  if (fBtnDown in [1,3]) then begin
    fTimer:= TTimer.Create(Self);
    fTimer.OnTimer:= TimerProc;
    fTimer.Interval:= 200;
    fTimer.Tag:= 5;
    TimerProc(fTimer);
    fTimer.Enabled:= True;
  end else
  if (fBtnDown=2) then begin
    SetCapture(Handle);
    fLastDragPos:= Y;
  end;
end;

procedure TPJSpin.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  inherited;
  fBtnDown:= 0;
  fTimer.Free;
  fTimer:= nil;
  Invalidate;
end;

//------------------------------------- painting ----------

procedure TPJSpin.Paint;
var s: string;
    w,h: Integer;
  procedure btnColor(i:Integer);
  begin
    if (i=fBtnActive) and (fBtnDown in [0,fBtnActive])
     then Canvas.Pen.Color:= fColors.ActiveBtn
     else Canvas.Pen.Color:= fColors.Btn;
    Canvas.Brush.Color:= Canvas.Pen.Color;
  end;
begin
  with Canvas do begin
    Font:= Self.Font;
    if fMouseIn or (fBtnDown<>0) then begin
      Brush.Color:= fColors.ActiveBack;
      Font.Color:= fColors.ActiveText;
    end
    else begin
      Brush.Color:= fColors.Back;
      Font.Color:= fColors.Text;
    end;
    s:= FloatToStrF(fValue, ffFixed, 7, fDigits);
    w:= TextWidth(s);
    h:= TextHeight(s);
    with fRects.Text do
     TextRect(ClipRect, Right-w, Top+(Bottom-Top-h) div 2, s);

    if Enabled and (fMouseIn or (fBtnDown<>0) or (csDesigning in ComponentState))
    then begin
      btnColor(1);
      with fRects.BtnUp do
       Polygon([
         Point(Left, Bottom-1),
         Point(Left + (Right-Left) div 2, Top),
         Point(Right, Bottom-1)
        ]);
      btnColor(2);
      with fRects.BtnUD do
       FillRect( Rect(Left, Top, succ(Right), succ(Bottom)) );
      btnColor(3);
      with fRects.BtnDn do
       Polygon([
         Point(Left, Top+1),
         Point(Left + (Right-Left) div 2, Bottom),
         Point(Right, Top+1)
        ]);
    end;
  end;
end;

//------------------------------------- .......... --------
procedure TPJSpin.TimerProc;
begin
  case fBtnDown of
    1 : if Assigned(fOnBtnUp)
         then fOnBtnUp(Self)
         else Value:= Value+fStep;
    3 : if Assigned(fOnBtnDn)
         then fOnBtnDn(Self)
         else Value:= Value-fStep;
  end;
  with TTimer(Sender) do
   if Tag>0 then begin
     Tag:= pred(Tag);
     if Tag=0 then begin
       if Interval=200 then Interval:= 100 else
       if Interval=100 then Interval:= 50  else
       if Interval=50  then Interval:= 10
       else Exit;
       Tag:= 10;
     end;
   end;
end;

//------------------------------------- .......... --------

end.
