unit katy_eulera;

interface

type
  Tos = (osX,osY,osZ);
  T3osie = array[1..3]of Tos;
  Tkombinacja = 1..12;

const
  KombinacjeEulera : array[Tkombinacja]of T3osie =
   { 1} ((osZ, osY, osZ),
   { 2}  (osZ, osX, osZ),
   { 3}  (osX, osY, osX),
   { 4}  (osX, osZ, osX),
   { 5}  (osY, osX, osY),
   { 6}  (osY, osZ, osY),
   { 7}  (osX, osY, osZ),
   { 8}  (osY, osZ, osX),
   { 9}  (osZ, osX, osY),
   {10}  (osX, osZ, osY),
   {11}  (osZ, osY, osX),
   {12}  (osY, osX, osZ));


procedure PoliczKatyEulera( nr_kombinacji: Tkombinacja;
                            wg_starych_osi: Boolean;
                            nx,ny,nz,
                            ox,oy,oz,
                            ax,ay,az: Single;
                            out a1,b1,c1,a2,b2,c2: Single );

function OpisKombinacji(kombinacja:Tkombinacja; wgstarych:Boolean):string;

//===================  definicje typw  =================//
type
  Tilosc = -1..1;
  Twartosc = string[2];
  Tfunkcja = (brak, sinus, cosinus, zlozona);
  Tkat = (kat1,kat2,kat3);

  // podstawowy element zwyklej macierzy
  //  ilosc   - -1 = wartosc ujemna
  //             0 = zero
  //             1 = wartosc dodatnia
  //  wartosc - dla macierzy T zawiera 'nx','ny'...
  //            dla wszystkich innych jest pusta ('')
  //  funkcja - czy w elemencie wystpuje jakas funkcja
  //            'zlozona' oznacza iloczyn dwch funkcji
  //  kat     - jesli jest funkcja - tutaj podane jest z jakiego kata
  //            gdy funkcja=brak - zmienna ta jest ignorowana
  Telement1 = record
    ilosc   : Tilosc;
    wartosc : Twartosc;
    funkcja : Tfunkcja;
    kat     : Tkat;
  end;

  // element macierzy powstalej w wyniku mnozenia dwoch innych
  // - moze byc sum dwch "zwyklych" elementw
  // (w przypadku uzywanych przez nas kombinacji macierzy nie takiej mozliwosci,
  // zeby powstala suma trzech elementow - przynajmniej jeden jest zawsze zerowy)
  Telement2 = array[1..2]of Telement1;

  Tmacierz1 = array[1..3,1..3]of Telement1;             // macierz "zwykla"
  Tmacierz2 = array[1..3,1..3]of Telement2;             // macierz wymnozona

  Twzor = array[1..2]of Telement2;    // wynik = arctg2( Twzor[1], Twzor[2] )


var
  wzory : array[Tkombinacja,Boolean,Tkat]of Twzor;


implementation

uses
  SysUtils, Math;


function OpisKombinacji(kombinacja:Tkombinacja; wgstarych:Boolean):string;
const znak : array[Tos]of char = ('X','Y','Z');
var osie : T3osie;
begin
  osie:= KombinacjeEulera[kombinacja];
  Result:= znak[osie[1]] +' '+ znak[osie[2]] +' '+ znak[osie[3]] +' ';
  if not wgstarych then begin
    Result[2]:= '''';
    Result[4]:= '''';
    Result[6]:= '''';
  end;
end;


//============ definicje podstawowych macierzy ==========//

//----- macierz kosinusow kierunkowych (nx,ny...) -----//
function macierzT:Tmacierz1;
var w,k:Integer;
begin
  for w:= 1 to 3 do
   for k:= 1 to 3 do
    with Result[w,k] do begin
      ilosc:= 1;
      funkcja:= brak;
      SetLength(wartosc, 2);
      case k of
        1 : wartosc[1]:= 'n';
        2 : wartosc[1]:= 'o';
        3 : wartosc[1]:= 'a';
      end;
      case w of
        1 : wartosc[2]:= 'x';
        2 : wartosc[2]:= 'y';
        3 : wartosc[2]:= 'z';
      end;
    end;
end;

//----- macierz obrotu wokol danej osi o dany kat -----//
function obrot(os:Tos; kat:Tkat):Tmacierz1;
  procedure el(w,k:Integer; ilosc:Tilosc; funkcja:Tfunkcja);
  begin
    Result[w,k].ilosc:= ilosc;
    Result[w,k].wartosc:= '';
    Result[w,k].funkcja:= funkcja;
    Result[w,k].kat:= kat;
  end;
begin
  case os of
    osX : begin
            el(1,1,1,brak);      el(1,2,0,brak);      el(1,3,0,brak);
            el(2,1,0,brak);      el(2,2,1,cosinus);   el(2,3,-1,sinus);
            el(3,1,0,brak);      el(3,2,1,sinus);     el(3,3,1,cosinus);
          end;
    osY : begin
            el(1,1,1,cosinus);   el(1,2,0,brak);      el(1,3,1,sinus);
            el(2,1,0,brak);      el(2,2,1,brak);      el(2,3,0,brak);
            el(3,1,-1,sinus);    el(3,2,0,brak);      el(3,3,1,cosinus);
          end;
    osZ : begin
            el(1,1,1,cosinus);   el(1,2,-1,sinus);    el(1,3,0,brak);
            el(2,1,1,sinus);     el(2,2,1,cosinus);   el(2,3,0,brak);
            el(3,1,0,brak);      el(3,2,0,brak);      el(3,3,1,brak);
          end;
  end;
end;


//================== mnozenie macierzy ==================//

   // ponisze funkcje zawieraj pewne uproszczenia
   // i raczej nie nadaj si do mnoenia innych
   // macierzy, ni te wystpujce przy wyznaczaniu
   // ktw Eulera.

function mnozenie_elem(a,b:Telement1):Telement1;
begin
  with Result do begin
    ilosc:= a.ilosc * b.ilosc;
    wartosc:= '';
    funkcja:= brak;
    if ilosc<>0 then begin
      if (a.funkcja=brak) then                           // a - zwykla wartosc
        if (b.funkcja=brak) then begin                          // b - zwykla wartosc
          wartosc:= a.wartosc + b.wartosc;
          // jedna z wartosci jest na pewno pusta
        end
        else begin                                              // b - funkcja
          wartosc:= a.wartosc;
          funkcja:= b.funkcja;
          kat:= b.kat;
        end
      else                                               // a - funkcja
        if (b.funkcja=brak) then begin                          // b - zwykla wartosc
          wartosc:= b.wartosc;
          funkcja:= a.funkcja;
          kat:= a.kat;
        end
        else begin                                              // b - funkcja
          funkcja:= zlozona;
          // taki element nas dalej nie interesuje
          // (za skomplikowany :))
        end;
    end;
  end;
end;

function mnozenie_wierszkol(m1,m2:Tmacierz1; w,k:Integer):Telement2;
var a,b,c: Telement1;
begin
  a:= mnozenie_elem(m1[w,1], m2[1,k]);
  b:= mnozenie_elem(m1[w,2], m2[2,k]);
  c:= mnozenie_elem(m1[w,3], m2[3,k]);
  // przynajmniej jedno z a,b,c jest na pewno zerowe i mozemy je opuscic
  if (a.ilosc=0) then begin  Result[1]:= b;  Result[2]:= c;  end else
  if (b.ilosc=0) then begin  Result[1]:= a;  Result[2]:= c;  end else
                      begin  Result[1]:= a;  Result[2]:= b;  end;
  // jesli tylko jedno bylo rozne od zera - wstawiamy je na 1. miejsce
  if (Result[1].ilosc=0) and (Result[2].ilosc<>0) then begin
    Result[1]:= Result[2];
    Result[2].ilosc:= 0;
  end;
end;


function mnozenie_macierzy(m1,m2:Tmacierz1):Tmacierz2;
var w,k: Integer;
begin
  for w:= 1 to 3 do
   for k:= 1 to 3 do
    Result[w,k]:= mnozenie_wierszkol(m1,m2,w,k);
end;


function macierz_odwrotna(m:Tmacierz1):Tmacierz1;
var w,k: Integer;
begin
  for w:= 1 to 3 do
   for k:= 1 to 3 do
    Result[w,k]:= m[k,w];
end;



//=================  wyznaczanie katw  =================//

// 'osie' i 'wg_starego' okrelaj pewn kombinacj
// wartoci k1, k2 i k3 zwracaj wzory na obliczenie kolejnych ktw
//  np. 'kat1 = arctg2( k1[1], k1[2] )'
procedure podaj_wzory(osie:T3osie; wg_starego:Boolean; out k1,k2,k3:Twzor);
var T,A,B,C: Tmacierz1;
    L,P: Tmacierz2;

  function szukaj_zera:Telement2;
  var w,k: Integer;
  begin
    for w:= 1 to 3 do
    for k:= 1 to 3 do
      if (P[w,k][2].ilosc=0) then
      if (P[w,k][1].ilosc=0) then begin
        Result:= L[w,k];
        Exit;
      end;
    raise Exception.Create('BD - szukaj_zera');
  end;

  function szukaj_funkcji(f:Tfunkcja; kt:Tkat):Telement2;
  var w,k: Integer;
  begin
    for w:= 1 to 3 do
    for k:= 1 to 3 do
      if (P[w,k][2].ilosc=0) then
      with P[w,k][1] do
       if (ilosc<>0) and (funkcja=f) and (kat=kt)
       then begin
         Result:= L[w,k];
         if ilosc=-1 then begin
           Result[1].ilosc:= -Result[1].ilosc;
           Result[2].ilosc:= -Result[2].ilosc;
         end;
         Exit;
       end;
    raise Exception.Create('BD - szukaj_funkcji');
  end;

  // wyznacza kat z zaleznosci typu: 'nx*c1 + ny*s1 = 0'
  //  pod 'el' podstawiam lew stron tego rwnania
  function rozbij(el:Telement2):Twzor;
   function r(i:Tilosc; w:Twartosc):Telement2;
   begin
     Result[1].ilosc:= i;
     Result[1].wartosc:= w;
     Result[1].funkcja:= brak;
     Result[2].ilosc:= 0;
   end;
  begin
    case el[1].funkcja of
      sinus   : begin
                  Result[1]:= r(-el[2].ilosc, el[2].wartosc);
                  Result[2]:= r( el[1].ilosc, el[1].wartosc);
                end;
      cosinus : begin
                  Result[1]:= r(-el[1].ilosc, el[1].wartosc);
                  Result[2]:= r( el[2].ilosc, el[2].wartosc);
                end;
    end;
    if Result[2][1].ilosc=-1 then begin
      { lepiej wyglda, jak w mianowniku nie ma minusa :)) }
      Result[1][1].ilosc:= -Result[1][1].ilosc;
      Result[2][1].ilosc:= 1;
    end;
  end;

begin
  (***** przygotowanie macierzy *****)
  T:= macierzT;
  A:= obrot(osie[1], kat1);
  B:= obrot(osie[2], kat2);
  C:= obrot(osie[3], kat3);

  (***** wymnoenie macierzy *****)
  // L - lewa strona
  // P - prawa strona
  if wg_starego then begin
    L:= mnozenie_macierzy(T, macierz_odwrotna(A));
    P:= mnozenie_macierzy(C, B);
  end
  else begin
    L:= mnozenie_macierzy(macierz_odwrotna(A), T);
    P:= mnozenie_macierzy(B, C);
  end;

  (***** porwnanie stron - L=P *****)
  k1:= rozbij(szukaj_zera);
  k2[1]:= szukaj_funkcji(sinus,kat2);
  k2[2]:= szukaj_funkcji(cosinus,kat2);
  k3[1]:= szukaj_funkcji(sinus,kat3);
  k3[2]:= szukaj_funkcji(cosinus,kat3);
end;


//==================  obliczanie katw  =================//

// tablica 'wzory' zawierajca wzory dla wszystkich
// mozliwych kombinacji przygotowywana jest raz
// przy starcie programu
procedure PrzygotowanieWzorow;
var i: Integer;
begin
  for i:= 1 to 12 do begin
    podaj_wzory(KombinacjeEulera[i], False,
      wzory[i, False, kat1],
      wzory[i, False, kat2],
      wzory[i, False, kat3]);
    podaj_wzory(KombinacjeEulera[i], True,
      wzory[i, True, kat1],
      wzory[i, True, kat2],
      wzory[i, True, kat3]);
  end;
end;

procedure PoliczKatyEulera;
var a: Single;
  function policz_el(el:Telement1):Single;
  begin
    with el do begin
      if (wartosc='nx') then Result:= nx else
      if (wartosc='ny') then Result:= ny else
      if (wartosc='nz') then Result:= nz else
      if (wartosc='ox') then Result:= ox else
      if (wartosc='oy') then Result:= oy else
      if (wartosc='oz') then Result:= oz else
      if (wartosc='ax') then Result:= ax else
      if (wartosc='ay') then Result:= ay else
      if (wartosc='az') then Result:= az else Result:= 1;
      if funkcja in [sinus,cosinus] then begin
        case funkcja of
          sinus   : Result:= Result * Sin(a);
          cosinus : Result:= Result * Cos(a);
        end;
      end;
      Result:= Result * ilosc;
    end;
  end;

  function policz_el2(el:Telement2):Single;
  begin
    Result:= policz_el(el[1]) + policz_el(el[2]);
  end;

  procedure policz_a(wzor:Twzor; var k1,k2:Single);
  var y,x: Single;
  begin
    y:= policz_el2(wzor[1]);
    x:= policz_el2(wzor[2]);
    k1:= ArcTan2( y, x);
    k2:= ArcTan2(-y,-x);
  end;

  procedure policz_bc(wzor:Twzor; var k1,k2:Single);
  begin
    a:= a1;  k1:= ArcTan2(policz_el2(wzor[1]), policz_el2(wzor[2]));
    a:= a2;  k2:= ArcTan2(policz_el2(wzor[1]), policz_el2(wzor[2]));
  end;

begin
  policz_a( wzory[nr_kombinacji, wg_starych_osi, kat1], a1, a2);
  policz_bc(wzory[nr_kombinacji, wg_starych_osi, kat2], b1, b2);
  policz_bc(wzory[nr_kombinacji, wg_starych_osi, kat3], c1, c2);
end;



initialization
  PrzygotowanieWzorow;

end.
